/* GDK - The GIMP Drawing Kit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/.
 */

#include "config.h"

#include "gdkprivate-directfb.h"

#include "gdkkeysyms.h"
#include "gdkinternals.h"

#include "gdkinput-directfb.h"


/*********************************************
 * Functions for maintaining the event queue *
 *********************************************/

static gboolean   dfb_events_prepare        (GSource         *source,
                                             gint            *timeout);
static gboolean   dfb_events_check          (GSource         *source);
static gboolean   dfb_events_dispatch       (GSource         *source,
                                             GSourceFunc      callback,
                                             gpointer         user_data);

static void       gdk_directfb_event_thread (gpointer         value);

static GdkEvent * gdk_event_translate       (DFBWindowEvent  *dfbevent,
                                             GdkWindow       *window);

static GList *event_windows = NULL;

static GSourceFuncs dfb_events_funcs =
  {
    dfb_events_prepare,
    dfb_events_check,
    dfb_events_dispatch,
    NULL
  };


guint32
gdk_directfb_get_time (void)
{
  GTimeVal tv;

  g_get_current_time (&tv);

  return (guint32) tv.tv_sec * 1000 + tv.tv_usec / 1000;
}


void
gdk_events_init (void)
{
  GSource *source;

  source = g_source_new (&dfb_events_funcs, sizeof (GSource));
  g_source_set_priority (source, GDK_PRIORITY_EVENTS);

  g_source_set_can_recurse (source, TRUE);
  g_source_attach (source, NULL);
}

/*
 *--------------------------------------------------------------
 * gdk_events_pending
 *
 *   Returns if events are pending on the queue.
 *
 * Arguments:
 *
 * Results:
 *   Returns TRUE if events are pending
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */

gboolean
gdk_events_pending (void)
{
  return gdk_event_queue_find_first () ? TRUE : FALSE;
}

GdkEvent *
gdk_event_get_graphics_expose (GdkWindow *window)
{
  GList *ltmp;

  g_return_val_if_fail (window != NULL, NULL);

  for (ltmp = gdk_queued_events; ltmp; ltmp = ltmp->next)
    {
      GdkEvent *event = ltmp->data;
      if (event->type == GDK_EXPOSE && event->expose.window == window)
        break;
    }

  if (ltmp)
    {
      GdkEvent *retval = ltmp->data;

      gdk_event_queue_remove_link (ltmp);
      g_list_free_1 (ltmp);

      return retval;
    }

  return NULL;
}

void
gdk_events_queue (void)
{
}

static gboolean
dfb_events_prepare (GSource *source,
                    gint    *timeout)
{
  GList *l = event_windows;
  DFBWindowEvent event;
  gboolean retval = FALSE;

  while (l)
    {
      GList                 *next = l->next;
      GdkWindow             *window = GDK_WINDOW (l->data);

      if (!GDK_WINDOW_DESTROYED (window))
        {
          GdkWindowObject       *private = GDK_WINDOW_OBJECT (window);
          GdkWindowImplDirectFB *impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);
          IDirectFBWindow       *iwindow = impl->window;

          while (iwindow->GetEvent( iwindow, &event ) == DFB_OK)
            {
              if (event.type == DWET_CLOSE)
                {
                  gdk_window_unref (window);
                  event_windows = g_list_remove_link ( event_windows, l );
                  g_list_free_1 ( l );
                }

              gdk_event_translate( &event, window );

              retval = TRUE;
            }
        }
      else
        {
          gdk_window_unref (window);
          event_windows = g_list_remove_link ( event_windows, l );
          g_list_free_1 ( l );
        }

      l = next;
    }

  *timeout = retval ? 1 : 10;

  return retval;
}

static gboolean
dfb_events_check (GSource *source)
{
  gboolean retval;

//   GDK_THREADS_ENTER ();

  retval = (gdk_event_queue_find_first () != NULL);

//   GDK_THREADS_LEAVE ();

  return retval;
}

static gboolean
dfb_events_dispatch (GSource     *source,
                     GSourceFunc  callback,
                     gpointer     user_data)
{
  GdkEvent *event;

//   GDK_THREADS_ENTER ();

  while ((event = gdk_event_unqueue ()) != NULL)
    {
      if (gdk_event_func)
        (*gdk_event_func) (event, gdk_event_data);

      gdk_event_free (event);
    }

//   GDK_THREADS_LEAVE ();

  return TRUE;
}

void
gdk_flush (void)
{
}

/* Sends a ClientMessage to all toplevel client windows */
gboolean
gdk_event_send_client_message (GdkEvent *event,
                               guint32   xid)
{
  return FALSE;
}

void
gdk_event_send_clientmessage_toall (GdkEvent *event)
{
}

/*****/

static void
gdk_directfb_event_thread (gpointer value)
{
  IDirectFBWindow       *iwindow;
  GdkWindowObject       *private;
  GdkWindowImplDirectFB *impl;
  GdkWindow             *window;

  g_return_if_fail (value != NULL);
  g_return_if_fail (GDK_IS_WINDOW (value));

  window = GDK_WINDOW (value);

  private = GDK_WINDOW_OBJECT (window);
  impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);
  iwindow = impl->window;

  while (TRUE)
    {
      DFBWindowEvent event;

      iwindow->WaitForEvent( iwindow );

      GDK_THREADS_ENTER ();

      while (iwindow->GetEvent( iwindow, &event ) == DFB_OK)
        gdk_event_translate( &event, window );

      GDK_THREADS_LEAVE ();
    }
}

void
gdk_directfb_start_event_thread (GdkWindow *window)
{
  GdkWindowImplDirectFB *impl;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  impl = GDK_WINDOW_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (window)->impl);

  if (!impl->window)
    return;

//   g_thread_create (gdk_directfb_event_thread, window, 0,
//                    TRUE, TRUE, G_THREAD_PRIORITY_NORMAL, NULL );

  event_windows = g_list_prepend( event_windows, gdk_window_ref(window) );
}

GdkWindow *
gdk_directfb_child_at (GdkWindow *window,
                       gint      *winx,
                       gint      *winy)
{
  GdkWindowObject *private;
  GList           *list;

  g_return_val_if_fail (window != NULL, NULL);
  g_return_val_if_fail (GDK_IS_WINDOW (window), NULL);

  private = GDK_WINDOW_OBJECT (window);

  for (list = private->children; list; list = list->next)
    {
      GdkWindowObject *win = list->data;

      if (GDK_WINDOW_IS_MAPPED (win) &&
          *winx >= win->x  &&
          *winx <  win->x + GDK_DRAWABLE_IMPL_DIRECTFB (win->impl)->width  &&
          *winy >= win->y  &&
          *winy <  win->y + GDK_DRAWABLE_IMPL_DIRECTFB (win->impl)->height)
        {
          *winx -= win->x;
          *winy -= win->y;

          return gdk_directfb_child_at (GDK_WINDOW (win), winx, winy );
        }
    }

  return window;
}

static GdkEvent *
gdk_event_translate (DFBWindowEvent *dfbevent,
                     GdkWindow      *window)
{
  GdkEvent        *event    = NULL;
  GdkWindowObject *private;

  g_return_val_if_fail (dfbevent != NULL, NULL);

  g_return_val_if_fail (window != NULL, NULL);
  g_return_val_if_fail (GDK_IS_WINDOW (window), NULL);

  private = GDK_WINDOW_OBJECT (window);

  switch (dfbevent->type)
    {
    case DWET_BUTTONDOWN:
    case DWET_BUTTONUP:
      {
        static int click_grab = 0;
        GdkWindow *child;
        guint      mask = 0;
        guint      button = 0;

        _gdk_directfb_mouse_x = dfbevent->cx;
        _gdk_directfb_mouse_y = dfbevent->cy;

        if (_gdk_directfb_pointer_grab_window) {
          GdkDrawableImplDirectFB *impl;

          child = _gdk_directfb_pointer_grab_window;
          impl = GDK_DRAWABLE_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (child)->impl);

          dfbevent->x = dfbevent->cx - impl->abs_x;
          dfbevent->y = dfbevent->cy - impl->abs_y;
        }
        else
          child = gdk_directfb_child_at( window, &dfbevent->x, &dfbevent->y );

        switch (dfbevent->button)
          {
          case DIBI_LEFT:
            button = 1;
            mask = GDK_BUTTON1_MASK;
            break;
          case DIBI_MIDDLE:
            button = 2;
            mask = GDK_BUTTON2_MASK;
            break;
          case DIBI_RIGHT:
            button = 3;
            mask = GDK_BUTTON3_MASK;
            break;
          }

        if (dfbevent->type == DWET_BUTTONDOWN)
          _gdk_directfb_modifiers |= mask;
        else
          _gdk_directfb_modifiers &= ~mask;

        event = gdk_event_make( child, dfbevent->type == DWET_BUTTONDOWN ?
                                GDK_BUTTON_PRESS : GDK_BUTTON_RELEASE, FALSE );

        event->button.x_root = _gdk_directfb_mouse_x;
        event->button.y_root = _gdk_directfb_mouse_y;

        event->button.x = dfbevent->x;
        event->button.y = dfbevent->y;

        event->button.state = _gdk_directfb_modifiers;
        event->button.button = button;
        event->button.device = gdk_core_pointer;

        g_message("button: %d at %d, %d with state 0x%08x",
                  event->button.button, (int)event->button.x,
                  (int)event->button.y, _gdk_directfb_modifiers);

        gdk_event_queue_append (event);

        if (dfbevent->type == DWET_BUTTONDOWN)
          gdk_event_button_generate (event);

         /* Handle implicit button grabs: */
        if (dfbevent->type == DWET_BUTTONDOWN  &&  !click_grab)
          {
            gdk_directfb_pointer_grab (child, FALSE,
                                       gdk_window_get_events (child),
                                       NULL, NULL,
                                       GDK_CURRENT_TIME, TRUE);
            click_grab = TRUE;
          }
        else if (dfbevent->type == DWET_BUTTONUP &&
                 !(_gdk_directfb_modifiers & (GDK_BUTTON1_MASK |
                                              GDK_BUTTON2_MASK |
                                              GDK_BUTTON3_MASK)) && click_grab)
          {
            gdk_directfb_pointer_ungrab (GDK_CURRENT_TIME, TRUE);
            click_grab = FALSE;
          }

        break;
      }

    case DWET_MOTION:
      {
        GdkWindow *child;

        _gdk_directfb_mouse_x = dfbevent->cx;
        _gdk_directfb_mouse_y = dfbevent->cy;

        if (_gdk_directfb_pointer_grab_window)
          {
            GdkDrawableImplDirectFB *impl;

            child = _gdk_directfb_pointer_grab_window;
            impl = GDK_DRAWABLE_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (child)->impl);

            dfbevent->x = dfbevent->cx - impl->abs_x;
            dfbevent->y = dfbevent->cy - impl->abs_y;
          }
        else
          child = gdk_directfb_child_at( window, &dfbevent->x, &dfbevent->y );

        event = gdk_event_make( child, GDK_MOTION_NOTIFY, TRUE );

        event->motion.x_root = _gdk_directfb_mouse_x;
        event->motion.y_root = _gdk_directfb_mouse_y;

        event->motion.x = dfbevent->x;
        event->motion.y = dfbevent->y;

        event->motion.state = _gdk_directfb_modifiers;
        event->motion.device = gdk_core_pointer;


        gdk_directfb_window_send_crossing_events (
                                                  NULL,
                                                  gdk_window_at_pointer (NULL,NULL),
                                                  GDK_CROSSING_NORMAL);

        break;
      }

    case DWET_GOTFOCUS:
      gdk_directfb_change_focus( window );

      break;

    case DWET_LOSTFOCUS:
      gdk_directfb_change_focus( gdk_parent_root );

      break;
      
    case DWET_POSITION_SIZE:
      {
        GdkWindow *event_win;

        private->x = dfbevent->x;
        private->y = dfbevent->y;

        GDK_DRAWABLE_IMPL_DIRECTFB (private->impl)->width = dfbevent->w;
	GDK_DRAWABLE_IMPL_DIRECTFB (private->impl)->height = dfbevent->h;

        event_win = gdk_directfb_other_event_window (window, GDK_CONFIGURE);
        if (event_win)
          {
            event = gdk_event_make (event_win, GDK_CONFIGURE, TRUE);
            event->configure.x = dfbevent->x;
            event->configure.y = dfbevent->y;
            event->configure.width = dfbevent->w;
            event->configure.height = dfbevent->h;
          }
        
        gdk_directfb_calc_abs( window );

	/* FIXME: if size changed do child re-sub-surfacing */
        
        break;
      }

    case DWET_POSITION:
      {
        GdkWindow *event_win;

        private->x = dfbevent->x;
        private->y = dfbevent->y;

        event_win = gdk_directfb_other_event_window (window, GDK_CONFIGURE);
        if (event_win)
          {
            event = gdk_event_make (event_win, GDK_CONFIGURE, TRUE);
            event->configure.x = dfbevent->x;
            event->configure.y = dfbevent->y;
            event->configure.width =
              GDK_DRAWABLE_IMPL_DIRECTFB (private->impl)->width;
            event->configure.height =
              GDK_DRAWABLE_IMPL_DIRECTFB (private->impl)->height;
          }
        
        gdk_directfb_calc_abs( window );
        
        break;
      }

    case DWET_SIZE:
      {
        GdkDrawableImplDirectFB *impl;
        GdkWindow               *event_win;
        GList                   *list;
        
        impl = GDK_DRAWABLE_IMPL_DIRECTFB (private->impl);

        event_win = gdk_directfb_other_event_window (window, GDK_CONFIGURE);
        if (event_win)
          {
            event = gdk_event_make (event_win, GDK_CONFIGURE, TRUE);
            event->configure.x      = private->x;
            event->configure.y      = private->y;
            event->configure.width  = dfbevent->w;
            event->configure.height = dfbevent->h;
          }

        impl->width  = dfbevent->w;
        impl->height = dfbevent->h;

        for (list = private->children; list; list = list->next)
          {
            GdkWindowObject         *win;
            GdkDrawableImplDirectFB *impl;

            win  = GDK_WINDOW_OBJECT (list->data);
            impl = GDK_DRAWABLE_IMPL_DIRECTFB (win->impl);

            gdk_directfb_move_resize_child (GDK_WINDOW (win),
                                            win->x, win->y,
                                            impl->width, impl->height);
          }

        break;
      }

    case DWET_KEYDOWN:
    case DWET_KEYUP:
      {
        GdkWindow *win =
          gdk_directfb_keyboard_event_window (gdk_directfb_window_find_focus (),
                                              dfbevent->type == DWET_KEYUP ?
                                              GDK_KEY_RELEASE : GDK_KEY_PRESS);
        if (win)
          {
            static guchar dummy[2] = { 0, 0 };
            guint  keycode;
            guchar keyascii;

            event = gdk_event_make (win, dfbevent->type == DWET_KEYUP ?
                                    GDK_KEY_RELEASE : GDK_KEY_PRESS, TRUE);

            gdk_directfb_translate_keycode (dfbevent, &keycode, &keyascii);
            
            dummy[0] = keyascii;

            gdk_directfb_convert_modifiers(dfbevent->modifiers);

            event->key.state            = _gdk_directfb_modifiers;
            event->key.keyval           = keycode;
            event->key.string           = g_strdup(dummy);
            event->key.length           = dummy[0] ? 1 : 0;
            event->key.hardware_keycode = dfbevent->keycode;
            event->key.group            = 0;
          }
        
        break;
      }

    case DWET_LEAVE:
      _gdk_directfb_mouse_x = dfbevent->cx;
      _gdk_directfb_mouse_y = dfbevent->cy;

      gdk_directfb_window_send_crossing_events (NULL, gdk_parent_root,
                                                GDK_CROSSING_NORMAL);

      if (gdk_directfb_apply_focus_opacity)
	{
	  if (GDK_WINDOW_IS_MAPPED (window))
	    GDK_WINDOW_IMPL_DIRECTFB (private->impl)->window->SetOpacity
	      (GDK_WINDOW_IMPL_DIRECTFB (private->impl)->window,
	       (GDK_WINDOW_IMPL_DIRECTFB (private->impl)->opacity >> 1) +
	       (GDK_WINDOW_IMPL_DIRECTFB (private->impl)->opacity >> 2));
	}

      break;

    case DWET_ENTER:
      {
        GdkWindow *child;

        _gdk_directfb_mouse_x = dfbevent->cx;
        _gdk_directfb_mouse_y = dfbevent->cy;

        child = gdk_directfb_child_at( window, &dfbevent->x, &dfbevent->y );

        gdk_directfb_window_send_crossing_events (NULL, child,
                                                  GDK_CROSSING_NORMAL);

      if (gdk_directfb_apply_focus_opacity)
	{
	  GDK_WINDOW_IMPL_DIRECTFB (private->impl)->window->SetOpacity
	    (GDK_WINDOW_IMPL_DIRECTFB (private->impl)->window,
	     GDK_WINDOW_IMPL_DIRECTFB (private->impl)->opacity);
	}

        break;
      }

    case DWET_CLOSE:
      {
        GdkWindow *event_win;

        event_win = gdk_directfb_other_event_window (window, GDK_DESTROY);
        if (event_win)
          {
            event = gdk_event_make (event_win, GDK_DESTROY, TRUE);
          }

	gdk_window_destroy_notify (window);

        break;
      }

    default:
      g_message( "unhandled DirectFB windowing event 0x%08x",
                 dfbevent->type);
    }

  return event;
}

gboolean
gdk_setting_get (const gchar *name,
                 GValue      *value)
{
  return FALSE;
}
