/* GDK - The GIMP Drawing Kit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 * Copyright (C) 1998-1999 Tor Lillqvist
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/.
 */

#include "config.h"

#include <stdlib.h>

#include "gdkpixmap.h"
#include "gdkinternals.h"
#include "gdkprivate-directfb.h"


static void gdk_pixmap_impl_directfb_init       (GdkPixmapImplDirectFB      *pixmap);
static void gdk_pixmap_impl_directfb_class_init (GdkPixmapImplDirectFBClass *klass);
static void gdk_pixmap_impl_directfb_finalize   (GObject                    *object);


static gpointer parent_class = NULL;


GType
gdk_pixmap_impl_directfb_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
        {
          sizeof (GdkPixmapImplDirectFBClass),
          (GBaseInitFunc) NULL,
          (GBaseFinalizeFunc) NULL,
          (GClassInitFunc) gdk_pixmap_impl_directfb_class_init,
          NULL,           /* class_finalize */
          NULL,           /* class_data */
          sizeof (GdkPixmapImplDirectFB),
          0,              /* n_preallocs */
          (GInstanceInitFunc) gdk_pixmap_impl_directfb_init,
        };
      
      object_type = g_type_register_static (GDK_TYPE_DRAWABLE_IMPL_DIRECTFB,
                                            "GdkPixmapImplDirectFB",
                                            &object_info, 0);
    }
  
  return object_type;
}

GType
_gdk_pixmap_impl_get_type (void)
{
  return gdk_pixmap_impl_directfb_get_type ();
}

static void
gdk_pixmap_impl_directfb_init (GdkPixmapImplDirectFB *impl)
{
  GdkDrawableImplDirectFB *draw_impl = GDK_DRAWABLE_IMPL_DIRECTFB (impl);
   
  draw_impl->width  = 1;
  draw_impl->height = 1;
}

static void
gdk_pixmap_impl_directfb_class_init (GdkPixmapImplDirectFBClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  object_class->finalize = gdk_pixmap_impl_directfb_finalize;
}

static void
gdk_pixmap_impl_directfb_finalize (GObject *object)
{
  GdkDrawableImplDirectFB *impl = GDK_DRAWABLE_IMPL_DIRECTFB (object);

  impl->surface->Release( impl->surface );
  free( impl->surface );

  if (G_OBJECT_CLASS (parent_class)->finalize)
    G_OBJECT_CLASS (parent_class)->finalize (object);
}

GdkPixmap*
gdk_pixmap_new (GdkWindow *window,
                gint       width,
                gint       height,
                gint       depth)
{
  IDirectFBSurface        *surface;
  GdkPixmap               *pixmap;
  GdkDrawableImplDirectFB *draw_impl;
  gboolean                 add_alpha = FALSE;

  g_return_val_if_fail (window == NULL || GDK_IS_WINDOW (window), NULL);
  g_return_val_if_fail ((window != NULL) || (depth != -1), NULL);
  g_return_val_if_fail ((width != 0) && (height != 0), NULL);

  if (!window)
    window = gdk_parent_root;

  if (GDK_WINDOW_DESTROYED (window))
    return NULL;

  GDK_NOTE (MISC, g_print ("gdk_pixmap_new: %dx%dx%d\n",
                           width, height, depth));
   
  if (depth == -1)
    {
      DFBSurfaceCapabilities caps;

      draw_impl = GDK_DRAWABLE_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (window)->impl);

      depth = BYTES_PER_PIXEL( draw_impl->format ) * 8;

      draw_impl->surface->GetCapabilities (draw_impl->surface, &caps);

      if (caps & DSCAPS_ALPHA)
	add_alpha = TRUE;
    }
  
  GDK_NOTE (MISC, g_print ("              : %dx%dx%d\n",
                           width, height, depth));
  switch (depth)
    {
    case  1:
    case  8:
    case 16:
    case 24:
    case 32:
      break;
    default:
      g_message("unimplemented %s for depth %d", __FUNCTION__, depth);
      return NULL;
    }

  pixmap = g_object_new (gdk_pixmap_get_type (), NULL);
  draw_impl = GDK_DRAWABLE_IMPL_DIRECTFB (GDK_PIXMAP_OBJECT (pixmap)->impl);

  {
    DFBResult             ret;
    DFBSurfaceDescription desc;

    desc.flags  = DSDESC_WIDTH | DSDESC_HEIGHT | DSDESC_BPP;
    desc.width  = width;
    desc.height = height;

    if (depth > 8)
      {
        desc.bpp = depth;

	if (add_alpha)
	  {
	    desc.flags |= DSDESC_CAPS;
	    desc.caps = DSCAPS_ALPHA;
	  }
      }
    else
      {
        desc.bpp = 8; /* FIXME */
        desc.flags |= DSDESC_CAPS;
        desc.caps = DSCAPS_ALPHA;
      }

    ret = DirectFB->CreateSurface( DirectFB, &desc, &surface );
    if (ret)
      {
        DirectFBError( "gdk_pixmap_new - CreateSurface", ret );
        return NULL;
      }
  }
   
  surface->GetSize( surface, &draw_impl->width, &draw_impl->height );
  surface->GetPixelFormat( surface, &draw_impl->format );

  draw_impl->abs_x = draw_impl->abs_y = 0;
   
  draw_impl->surface = surface;

  GDK_PIXMAP_OBJECT (pixmap)->depth = BITS_PER_PIXEL(draw_impl->format);

  return pixmap;
}

GdkPixmap *
gdk_bitmap_create_from_data (GdkWindow   *window,
                             const gchar *data,
                             gint         width,
                             gint         height)
{
  g_message("unimplemented %s", __FUNCTION__);

  return NULL;
}

GdkPixmap*
gdk_pixmap_create_from_data (GdkWindow   *window,
                             const gchar *data,
                             gint         width,
                             gint         height,
                             gint         depth,
                             GdkColor    *fg,
                             GdkColor    *bg)
{
  GdkPixmap *pixmap;

  GDK_NOTE (MISC, g_print ("gdk_pixmap_create_from_data: %dx%dx%d\n",
                           width, height, depth));
   
  pixmap = gdk_pixmap_new (window, width, height, depth);

  if (pixmap) 
    {
      gchar *dst;
      gint   pitch;
      gint   src_pitch = width * ((depth+7)/8);
      
      IDirectFBSurface *surface;
         
      surface = GDK_DRAWABLE_IMPL_DIRECTFB (GDK_PIXMAP_OBJECT (pixmap)->impl)->surface;

      if (surface->Lock( surface, DSLF_WRITE, (void**)(&dst), &pitch ) == DFB_OK)
        {
          gint i;

          for (i = 0; i < height; i++)
            {
              memcpy( dst, data, src_pitch );
              dst += pitch;
              data += src_pitch;
            }
          
          surface->Unlock( surface );
        }
    }

  return pixmap;
}

